import './redirect';
import type { ESLintConfig, Rules } from 'eslint-define-config';
import { typescriptRules } from './presets/typescript';
import { unicornRules } from './presets/unicorn';
import { eslintRules } from './presets/eslint';
import { reactRules } from './presets/react';
import { importRules } from './presets/import';
import { error, warn, off } from './constants';
// @ts-expect-error
const { name } = [require][0]('./package.json');

export { error, warn, off };

const unique = <T>(arr: T[]): T[] => [...new Set(arr)];
const ensureArray = <T>(value?: T | T[]): T[] =>
  value == null ? [] : Array.isArray(value) ? value : [value];

type RuleLevel = 'error' | 'warn' | 'off' | 0 | 1 | 2;
type RuleEntry<Options> = RuleLevel | [RuleLevel, Partial<Options>];

declare module 'eslint-define-config/src/rules/react/no-unknown-property.d.ts' {
  export interface NoUnknownPropertyOption {
    extends: ('next' | 'emotion')[];
  }
}

export interface LocalRuleOptions {
  /** Bans import from the specifier '.' and '..' and replaces it with '.+/index' */
  'rules/no-import-dot': RuleEntry<unknown>;
  /**
   * Enforce template literal expressions to be of `string` type
   * @see [restrict-template-expressions](https://typescript-eslint.io/rules/restrict-template-expressions)
   */
  'rules/restrict-template-expressions': RuleEntry<{ allow: string[] }>;
  /** Ban assignment of empty object literals `{}` and replace them with `Object.create(null)` */
  'rules/no-empty-object-literal': RuleEntry<unknown>;
}

export type RuleOptions = Rules & Partial<LocalRuleOptions>;

/**
 * ESLint Configuration.
 * @see [ESLint Configuration](https://eslint.org/docs/latest/user-guide/configuring/)
 */
type Config = Omit<ESLintConfig, 'rules'> & {
  /**
   * Rules.
   * @see [Rules](https://eslint.org/docs/latest/user-guide/configuring/rules)
   */
  rules?: RuleOptions;
};

/**
 * Returns a ESLint config object.
 *
 * By default, it includes `["@typescript-eslint", "import", "prettier"]` configs.
 * Additional bundled plugins include `["react", "react-refresh", "jsx-a11y", "unicorn"]`.
 */
export function extendConfig({
  plugins,
  settings,
  rules,
  extends: _extends,
  overrides,
  ...rest
}: Config = {}): ESLintConfig {
  const hasReact = plugins?.includes('react');
  const hasReactRefresh = plugins?.includes('react-refresh');
  const hasUnicorn = plugins?.includes('unicorn');
  const hasNext = ensureArray(_extends).some(name => name.includes(':@next/next'));

  const result: Config = {
    root: true,
    parser: '@typescript-eslint/parser',
    plugins: unique(['@typescript-eslint', 'import', 'rules', ...(plugins ?? [])]),
    env: { node: true, browser: true, es2023: true },
    reportUnusedDisableDirectives: true,
    parserOptions: {
      project: true,
    },
    extends: unique([
      'eslint:recommended',
      'prettier',
      'plugin:@typescript-eslint/recommended-type-checked',
      'plugin:import/errors',
      'plugin:import/typescript',
      ...(hasReact
        ? [
            'plugin:react/recommended',
            'plugin:react-hooks/recommended',
            'plugin:jsx-a11y/recommended',
          ]
        : []),
      ...(_extends ?? []),
    ]),
    settings: {
      'import/parsers': {
        '@typescript-eslint/parser': ['.ts', '.tsx', '.mts', '.cts'],
      },
      'import/resolver': {
        typescript: {
          alwaysTryTypes: true,
        },
      },
      react: {
        version: 'detect',
      },
      ...settings,
    },
    overrides: [
      {
        files: [
          '.eslintrc.js',
          '*.config.js',
          'index.js',
          'babel.config.js',
          'next.config.js',
        ],
        extends: ['plugin:@typescript-eslint/disable-type-checked'],
        rules: {
          'rules/restrict-template-expressions': off,
        },
      },
      {
        files: ['*.d.ts'],
        rules: {
          '@typescript-eslint/consistent-type-imports': off,
        },
      },
      ...(overrides ?? []),
    ],
    rules: {
      ...eslintRules,
      ...typescriptRules,
      ...importRules,
      ...(hasReact && {
        ...reactRules,
        'react/no-unknown-property': [
          error,
          { ignore: hasNext ? ['emotion', 'next'] : ['emotion'] },
        ],
      }),
      ...(hasReactRefresh && {
        'react-refresh/only-export-components': [warn, { allowConstantExport: true }],
      }),
      ...(hasUnicorn && unicornRules),
      'rules/no-import-dot': error,
      'rules/restrict-template-expressions': error,
      ...rules,
    },
    ...rest,
  };

  return result;
}
